/***************************************************************************
 *
 *                                  FROST                  
 *                Forensic Recovery Of Scrambled Telephones
 *  This is a kernel module to recover AES keys from RAM of Android phones.
 *
 * Copyright (C) 2012 Tilo Mueller <tilo.mueller@informatik.uni-erlangen.de>
 *
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 *
 * 
 * The AES key recovery functionality is based on AESKeyFinder (2008-07-18).
 * http://citp.princeton.edu/research/memory/code by N. Heninger and A. Feldman
 *
 ***************************************************************************/

#include <linux/highmem.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/pfn.h>

int fullsearch = 0;				/* 0 = quick search  1 = full search */
module_param(fullsearch, int, S_IRUGO);

#define SEARCH_SPACE_START (void *)0xc5000000	/* where to begin in quick search mode */
#define SEARCH_SPACE_END (void *)0xd0000000	/* where to end in quick search mode */
#define AESKEYFIND_THRESHOLD 64			/* threshold for AES key schedules (bits) */
#define STRINGDIFF_THRESHOLD 5			/* threshold for magic strings (bytes) */

extern struct resource iomem_resource;		/* start point of physical page walkthrough */

uint8_t BMAP[PAGE_SIZE*2];			/* sliding window in full search mode */

uint8_t sbox[256] = {
	0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
	0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
	0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
	0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
	0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
	0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
	0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
	0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
	0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
	0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
	0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
	0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
	0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
	0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
	0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
	0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16 };

uint8_t rcon[255] = {
	0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 
	0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 
	0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 
	0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 
	0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 
	0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 
	0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 
	0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 
	0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 
	0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 
	0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 
	0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 0x61, 0xc2, 0x9f, 
	0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb, 0x8d, 0x01, 0x02, 0x04, 
	0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 
	0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91, 0x39, 0x72, 0xe4, 0xd3, 0xbd, 
	0x61, 0xc2, 0x9f, 0x25, 0x4a, 0x94, 0x33, 0x66, 0xcc, 0x83, 0x1d, 0x3a, 0x74, 0xe8, 0xcb };

/* Set byte n of vector to val. */
static inline uint32_t set_byte(uint32_t vector, int n, uint8_t val)
{
	return (vector & ~(0xFF << (8*n))) | (val << (8*n));
}

/* Return byte n of vector. */
static inline uint8_t get_byte(uint32_t vector, int n)
{
	return (vector >> (8*n)) & 0xFF;
}

/* Perform the AES key core operation on a word.
 * (reverts the byte order to fix some weird endianness bugs) */
static inline uint32_t key_core(uint32_t k, int i)
{
	uint32_t k2 = 0,t = 0,t2 = 0;
	int j;
	k2 = ((k&0xff)<<24 ^ (k&0xff00)<<8 ^ (k&0xff0000)>>8 ^ (k&0xff000000)>>24);
	for (j=0; j<4; j++)
		t = set_byte(t, j, sbox[get_byte(k2,j)]);
	t = set_byte(t, 2, get_byte(t,2) ^ rcon[i]);
	t2 = ((t&0xff)<<16 ^ (t&0xff0000)>>16 ^ (t&0xff00ff00));
	return t2;
}

/* Run each byte of a word through the sbox separately for word 4 of 256-bit AES.
 * (reverts the byte order to fix some weird endianness bugs) */
static inline uint32_t sbox_bytes(uint32_t k)
{
	uint32_t r = 0, r2 = 0, k2 = 0;
	int j;
	k2 = ((k&0xff)<<24 ^ (k&0xff00)<<8 ^ (k&0xff0000)>>8 ^ (k&0xff000000)>>24);
	for (j=0; j<4; j++)
		r = set_byte(r,j,sbox[get_byte(k2,j)]);
	r2 = ((r&0xff)<<24 ^ (r&0xff00)<<8 ^ (r&0xff0000)>>8 ^ (r&0xff000000)>>24);
	return r2;
}

/* Return the number of bits in x that are 1. */
static inline int popcount(uint64_t x)
{
	int count=0,i;
	for (i=0; i<64; i++) count += (x>>i)&1;
	return count;
}

/* Print the AES key found. */
static void print_key(uint8_t* map)
{
	int i;
	printk("[FROST] key-16 found: ");
	for (i = 0; i < 16; i++)
		printk("%02x",map[i]);
	printk("\n");
}

/* Simple entropy test.
 * Returns true if the 176 bytes starting at location bmap[i] contain
 * more than 16 repeats of any byte. */
static inline int entropy(const uint8_t* bm, int i)
{
	static int byte_freq[256] = {0};
	int test = 0, b;
	if (i == 0) {
		int j; 
		for (j=0; j<256; j++) byte_freq[j] = 0;
		for (j=0; j<176; j++) byte_freq[bm[j]]++;
	}
	for (b=0; b<=0xFF; b++) {
		if (byte_freq[b] > 16) {
			test = 1;
			break;
		}
	}
	byte_freq[bm[i]]--;
	byte_freq[bm[i+176]]++;
	return test;
}

/* The core key finding loop.
 * Searches for AES-128 keys in memory image bmap with given size;
 * prints any keys found. */
int find_keys(const uint8_t *bmap, size_t size)
{
	size_t i; int res = 0;
	uint32_t *map;
	int xor_count_128; 
	size_t row, column; 

	for (i = 0; i < size-416; i+=16) {

		map = (uint32_t *)&(bmap[i]);
		if (entropy(bmap,i)) continue;

		xor_count_128 = 0;
		for (row = 1; row < 11; row++) {
			for (column = 0; column < 4; column++) {
				if (column == 0)
					xor_count_128 += popcount(key_core(map[4*row-1],row) ^
							map[4*(row-1)] ^
							map[4*row]);
				else
					xor_count_128 += popcount((map[4*row + column-1] ^
							map[4*(row-1)+column]) ^
							map[4*row + column]);
			}
			if (xor_count_128 > AESKEYFIND_THRESHOLD)
				break;
		}
		if (xor_count_128 <= AESKEYFIND_THRESHOLD) {
			print_key((uint8_t *)map);
			res++;
		}
	}

	return res;
}

/* A wrapper for the key finding function.
 * Searches for AES keys in memory image bmap0|bmap1 with size size0+size1. */
static inline int find_keys_p2(const uint8_t *bmap0, const uint8_t *bmap1, size_t size0, size_t size1)
{
	size_t i, found_keys = 0;

	for (i=0; i<size0; i++) BMAP[i] = bmap0[i];
	for (i=0; i<size1; i++) BMAP[size0+i] = bmap1[i];
	
	if ((found_keys = find_keys(BMAP,size0+size1)))
		printk("[FROST] Found key(s) at address %p - %p.\n",bmap0,bmap1+size1);

	return found_keys;
}

/* Compare constant string b with noisy byte array a.
 * Returns true if less the STRINGDIFF_THRESHOLD bytes between a and b differ. */
static inline int strdiff(const uint8_t *a, const char *b)
{
	int i, diffbytes=0;

	for (i=0; i<strlen(b); i++) {
		if (a[i]^b[i]) diffbytes++;
		if (diffbytes > STRINGDIFF_THRESHOLD) return 0;
	}

	return 1;
}

/* Physical page walkthrough.
 * In full search mode a sliding window mechanism over two phyiscal pages is
 * used. In quick search mode only single pages are considered (and thus AES
 * key schedules spread over two pages are missed). */
int init_module(void)
{
	int found_keys = 0;
	struct resource *res;

	for (res = iomem_resource.child; res ; res = res->sibling) {
		if (strncmp(res->name, "System RAM", sizeof("System RAM")))
			continue;
		else if (fullsearch) { /* full search (general code) */
			mm_segment_t fs;
			resource_size_t i;
			struct page *p0=NULL,*p1=NULL;
			void *v0=NULL,*v1=NULL;
			int turn=0, round=0;
			fs = get_fs();
			set_fs(KERNEL_DS);
			for (i = res->start; i < res->end; i += PAGE_SIZE) {
				if (turn) {
					p1 = pfn_to_page(PFN_DOWN(i));
					v1 = kmap(p1);
					if (round) {
						found_keys += find_keys_p2(v0,v1,PAGE_SIZE,PAGE_SIZE);
						kunmap(p0);
					}
				}
				else {
					p0 = pfn_to_page(PFN_DOWN(i));
					v0 = kmap(p0);
					if (round) {
						found_keys += find_keys_p2(v1,v0,PAGE_SIZE,PAGE_SIZE);
						kunmap(p1);
					}
				}
				turn = 1-turn;
				round++;
			}
			if (round == 1) found_keys += find_keys(v0,PAGE_SIZE);
			if (turn) kunmap(p0);
			else kunmap(p1);
			set_fs(fs);
		}
		else { /* quick search (Android/GNex specific) */
			mm_segment_t fs;
			resource_size_t i;
			struct page *p;
			void *v;
			fs = get_fs();
			set_fs(KERNEL_DS);
			for (i = res->start; i < res->end; i += PAGE_SIZE) {
				p = pfn_to_page(PFN_DOWN(i));
				v = kmap(p);
				if (v > SEARCH_SPACE_START && v < SEARCH_SPACE_END)
					found_keys += find_keys(v,PAGE_SIZE);
				kunmap(p);
			}
			set_fs(fs);
		}
	}

	/* No key schedule found (likely because the image is too noisy).
	 * Last chance is to recover (maybe disturbed) keys near magic strings.
	 * Should be highly Android and GNex specific.
	 */
	if (!found_keys) {
		printk("[FROST] Warning: No AES key schedules were found.\n");
		printk("[FROST] The following keys might be disturbed.\n");
		for (res = iomem_resource.child; res ; res = res->sibling) {
			if (strncmp(res->name, "System RAM", sizeof("System RAM")))
				continue;
			else {
				int j;
				mm_segment_t fs;
				resource_size_t i;
				struct page *p;
				uint8_t *v;
				fs = get_fs();
				set_fs(KERNEL_DS);
				for (i = res->start; i < res->end; i += PAGE_SIZE) {
					p = pfn_to_page(PFN_DOWN(i));
					v = (uint8_t *)kmap(p);
					for (j=32; j<PAGE_SIZE-128; j+=4) {
						if (strdiff(v+j,"/data/tmp_mnt") && strdiff(v+j+64,"encrypted")) {
							print_key(v+j-32);
							found_keys++;
						}
						if (strdiff(v+j,"/fstab.tuna") && ((v[j+120] == 0x01) || (v[j+120] == 0x41) || (v[j+120] == 0x81))) {
							print_key(v+j+104);
							found_keys++;
						}
					}
					kunmap(p);
				}
				set_fs(fs);
			}
		}
	}
		
	printk("[FROST] Summarizing %i keys found.\n",found_keys);
	return 0;
}

void cleanup_module(void) { }
MODULE_LICENSE("GPL");
MODULE_AUTHOR ("Tilo Mueller, tilo.mueller@informatik.uni-erlangen.de");
MODULE_DESCRIPTION ("AES key recovery from scrambled Android phones.");
